/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2016 KYOCERA Corporation
 */


#include "logcapture.h"

#include <linux/hrtimer.h>
#include <linux/kmod.h>


#define DUMPLOG_TIMER 30
#define LOG_KEY_NONE  0x0
#define LOG_KEY_VOLDN 0x02
#define LOG_KEY_VOLUP 0x01
#define LOG_KEY_ON    (LOG_KEY_VOLDN | LOG_KEY_VOLUP)

#ifdef CONFIG_LOGCAPTURE_DEBUG
#define LOGCAPTURE_PRINT(fmt, ...) printk(KERN_DEBUG fmt, ##__VA_ARGS__)
#else
#define LOGCAPTURE_PRINT(fmt, ...) do {} while (0)
#endif

static struct hrtimer logcapture_timer;
static int logcapture_status = LOG_KEY_NONE;

void check_logcapture(struct input_dev* dev,
                               unsigned int code, int value)
{
	if (value == 0) {
		if (code == KEY_VOLUMEUP) {
			LOGCAPTURE_PRINT("logcapture status=0x%x dev=%p\n",
				logcapture_status, dev);
			logcapture_status &= ~LOG_KEY_VOLUP;
			LOGCAPTURE_PRINT("logcapture timer cancel\n");
			hrtimer_cancel(&logcapture_timer);
			LOGCAPTURE_PRINT("logcapture after status=0x%x\n",
				logcapture_status);
		} else if (code == KEY_VOLUMEDOWN) {
			LOGCAPTURE_PRINT("logcapture status=0x%x dev=%p\n",
				logcapture_status, dev);
			logcapture_status &= ~LOG_KEY_VOLDN;
			LOGCAPTURE_PRINT("logcapture timer cancel\n");
			hrtimer_cancel(&logcapture_timer);
			LOGCAPTURE_PRINT("logcapture after status=0x%x\n",
				logcapture_status);
		}
	} else {
		if (code == KEY_VOLUMEUP) {
			LOGCAPTURE_PRINT("logcapture status=0x%x dev=%p\n",
				logcapture_status, dev);
			logcapture_status |= LOG_KEY_VOLUP;
			if (logcapture_status == LOG_KEY_ON) {
				LOGCAPTURE_PRINT("logcapture timer start\n");
				hrtimer_start(&logcapture_timer,
				ktime_set(DUMPLOG_TIMER,0),
				HRTIMER_MODE_REL);
			}
			LOGCAPTURE_PRINT("logcapture after status=0x%x\n",
				logcapture_status);
		} else if (code == KEY_VOLUMEDOWN) {
			LOGCAPTURE_PRINT("logcapture status=0x%x dev=%p \n",
				logcapture_status, dev);
			logcapture_status |= LOG_KEY_VOLDN;
			if (logcapture_status == LOG_KEY_ON) {
				LOGCAPTURE_PRINT("logcapture timer start\n");
				hrtimer_start(&(logcapture_timer),
					ktime_set(DUMPLOG_TIMER,0),
					HRTIMER_MODE_REL);
			}
			LOGCAPTURE_PRINT("logcapture after status=0x%x\n",
				logcapture_status);
		}
	}
}


static int execute_logcapture(void)
{
	static char *envp[] = {
		"HOME=/",
		"TERM=linux",
		"PATH=/sbin:/vendor/bin:/system/sbin:/system/bin:/system/xbin",
		NULL
	};
	static char *argv[] = {
		"/system/vendor/bin/logcapture",
		"--start",
		NULL
	};
	int ret;

	ret = call_usermodehelper(argv[0], argv, envp, UMH_NO_WAIT);
	LOGCAPTURE_PRINT("++ %s: after input call_usermodehelper() (ret=%d).\n",
		__func__, ret);
	if (ret < 0) {
		return ret;
	}
	return 0;
}

static enum hrtimer_restart logcapture_timer_func(struct hrtimer *timer)
{
	LOGCAPTURE_PRINT("logcapture_timer_func start\n");
	execute_logcapture();
	return HRTIMER_NORESTART;
}


/* call from input_init */
void __init logcapture_init(void)
{
	LOGCAPTURE_PRINT("logcapture_init\n");
	hrtimer_init(&logcapture_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	logcapture_timer.function = logcapture_timer_func;
}

/* call from input_exit */
void __exit logcapture_exit(void)
{
	LOGCAPTURE_PRINT("logcapture_exit\n");
	hrtimer_cancel(&logcapture_timer);
}
